
#define ENABLE_TIMER 0

#include <functional>
#include <algorithm>
#include <chrono>
#include <iostream>
#include <fstream>

#if ENABLE_TIMER
#include "hark-lib/node/strDurlog.hpp"
#endif

#include "hark-lib/node/LocalizationMUSIC.hpp"

#include <math.h>

////#ifdef ENABLE_HARKIO3
extern "C" {
#include <libharkio3.h>
}
////#endif // #ifdef ENABLE_HARKIO3

#include "hark-lib/matrix.h"

namespace hark
{

struct less_for_vec {
    bool operator()(std::vector<float>& v, int i, int j) {
        return v[i] < v[j];
    }
};

void AddToCorrelation(std::vector<Matrix<std::complex<float> > >& Rxx, const std::vector<bool>& UsesFreq, Matrix<std::complex<float> >& X)
{
    size_t k_max = std::min(Rxx.size(), UsesFreq.size());
    for(size_t k = 0; k < k_max; k++) {
        if(UsesFreq[k]) {
            Rxx[k] += X.col(k) *  X.col(k).adjoint();
            //Rxx[k] += outer_prod(ublas::column(X, k), ublas::herm(ublas::column(X, k)));
        }
    }
}

void NormalizeCorrelation(std::vector<Matrix<std::complex<float> > >& Rxx, const std::vector<bool>& UsesFreq, int sum_count)
{
    size_t k_max = std::min(Rxx.size(), UsesFreq.size());
    float sum_count_1 = 1.0f / sum_count;
    for(size_t k = 0; k < k_max; k++) {
        if(UsesFreq[k]) {
            Rxx[k] *= sum_count_1;
        }
    }
}

void ClearCorrelation(std::vector<Matrix<std::complex<float> > >& Rxx, const std::vector<bool>& UsesFreq)
{
    for(size_t k = 0; k < Rxx.size(); k++) {
        //for(size_t i = 0; i < Rxx[k].rows(); i++){
        //    for(size_t j = 0; j < Rxx[k].cols(); j++){
        //        Rxx[k](i, j) = 0.0f;
        //    }
        //}
        Rxx[k].setZero();
    }
}

float MaxOfAbsValue(Matrix<std::complex<float> >& A)
{
    float max = 0.0f;
    for(size_t i = 0; i < A.rows(); i++) {
        for(size_t j = 0; j < A.cols(); j++) {
            if(std::abs(A(i, j).real()) > max) {
                max = std::abs(A(i, j).real());
            }
        }
    }
    return max;
    //return A.real().cwiseAbs().maxCoeff(); // I have not checked.
    //return A.real().array().abs().maxCoeff(); // I have not checked.
}

////////////////////////////////////////////////////////////////////////////////
/*
float MaxOfAbsValueFvector(std::vector<float>& A)
{
    float max = 0.0f;
    for(size_t i = 0; i < A.size(); i++) {
        if(std::abs(A[i]) > max){
            max = std::abs(A[i]);
        }
    }
    return max;
}

void mulfcomplex(std::complex<float>& a, std::complex<float>& b, std::complex<float>& c)
{
    float tmp1, tmp2;
    tmp1 = a.real() * b.real() - a.imag() * b.imag();
    tmp2 = a.real() * b.imag() + a.imag() * b.real();
    c = std::complex<float>(tmp1, tmp2);
}

void addfcomplex(std::complex<float>& a, std::complex<float>& b, std::complex<float>& c)
{
    float tmp1, tmp2;
    tmp1 = a.real() + b.real();
    tmp2 = b.imag() + a.imag();
    c = std::complex<float>(tmp1, tmp2);
}

void subfcomplex(std::complex<float>& a, std::complex<float>& b, std::complex<float>& c)
{
    float tmp1, tmp2;
    tmp1 = a.real() - b.real();
    tmp2 = a.imag() - b.imag();
    c = std::complex<float>(tmp1, tmp2);
}

void devfcomplex(std::complex<float>& a, std::complex<float>& b, std::complex<float>& c)
{
    float tmp1, tmp2;
    tmp1 = (a.real() * b.real() + a.imag() * b.imag()) / (b.real() * b.real() + b.imag() * b.imag());
    tmp2 = (a.imag() * b.real() - a.real() * b.imag()) / (b.real() * b.real() + b.imag() * b.imag());
    c = std::complex<float>(tmp1, tmp2);
}

void mulfcomplex_const(std::complex<float> a, std::complex<float> b, std::complex<float>& c)
{
    mulfcomplex(a, b, c);
}

void addfcomplex_const(std::complex<float> a, std::complex<float> b, std::complex<float>& c)
{
    addfcomplex(a, b, c);
}

void subfcomplex_const(std::complex<float> a, std::complex<float> b, std::complex<float>& c)
{
    subfcomplex(a, b, c);
}

void devfcomplex_const(std::complex<float> a, std::complex<float> b, std::complex<float>& c)
{
    devfcomplex(a, b, c);
}
*/





LocalizationMUSIC::LocalizationMUSIC(int num_mic, int sampling_rate, int tlength, int shift, int maxn_out_peaks) :
    NumMicrophones(num_mic), TimeLength(tlength), Shift(shift), SamplingRate(sampling_rate), R(num_mic, num_mic), V(
        num_mic, num_mic), RN(num_mic, num_mic), VN(num_mic, num_mic), TempCalc1(num_mic, num_mic), TempCalc2(num_mic,
        num_mic), TempCalc3(num_mic, num_mic), TempCalc4(num_mic, num_mic), TempCalc5(num_mic, num_mic), TempCalc6(
        num_mic, num_mic), TempCalc7(num_mic, num_mic), TempCalc8(num_mic, num_mic), TempCalc9(num_mic, num_mic), NumSource(
        num_mic) {
    FreqLength = TimeLength / 2 + 1;

    UsesFreq.resize(FreqLength);

    maxnum_output_peaks = maxn_out_peaks;

    EigenValue.resize(FreqLength);
    CEigenValue.resize(NumMicrophones);
    REigenValue.resize(NumMicrophones);
    MaxRxx.resize(FreqLength);
    MaxRxxN.resize(FreqLength);
    E.resize(FreqLength);
    ////EN.resize(FreqLength);

    //for(size_t k = 0; k < FreqLength; k++) {
    //    EigenValue[k].resize(NumMicrophones);
    //    Matrix<std::complex<float> > Emat(NumMicrophones, NumMicrophones);
    //    Matrix<std::complex<float> > ENmat(NumMicrophones, NumMicrophones);
    //    E[k] = Emat;
    //    EN[k] = ENmat;
    //}
    for(size_t k = 0; k < FreqLength; k++) {
        EigenValue[k].resize(NumMicrophones);
        E[k].resize(NumMicrophones, NumMicrophones);
        ////EN[k].resize(NumMicrophones, NumMicrophones);
    }

    FoundId.resize(NumSource);
}

void LocalizationMUSIC::SetUsedFrequency(int lo, int up) {
    size_t lower = (int) ((float) lo * TimeLength / SamplingRate);
    size_t upper = (int) ((float) up * TimeLength / SamplingRate);
    for(size_t k = 0; k < lower; k++) {
        UsesFreq[k] = false;
    }
    for(size_t k = lower; k < upper; k++) {
        UsesFreq[k] = true;
    }
    for(size_t k = upper; k < UsesFreq.size(); k++) {
        UsesFreq[k] = false;
    }
}

void LocalizationMUSIC::Localize(std::vector<Matrix<std::complex<float> > >& Rxx,
                                 std::vector<Matrix<std::complex<float> > >& RxxN,
                                 int mindeg, int maxdeg,
                                 std::string music_alg, std::vector<float>& MSSW, bool enable_eigenvalue_weight) {
#if ENABLE_TIMER
    std::chrono::system_clock::time_point start_time;
    std::chrono::system_clock::time_point end_time;
    start_time = std::chrono::system_clock::now();
#endif

    for(size_t k = 0; k < UsesFreq.size(); k++) {

        MaxRxx[k] = 1.0;
        MaxRxxN[k] = 1.0;

        if(!UsesFreq[k]) {
            continue;
        }
        float max = MaxOfAbsValue(Rxx[k]);
        R = Rxx[k] / (max + 1.0f);
        MaxRxx[k] = max + 1.0f;

        if(music_alg != "SEVD") {
            float maxN = MaxOfAbsValue(RxxN[k]);
            RN = RxxN[k] / (maxN);
            MaxRxxN[k] = maxN;
        }
        //============= GEVD ==============
        //printf("%d ", k);

        if(music_alg == "SEVD") {
            //LAPACK with HARK_Matrix// Matrix<std::complex<float> > EV(NumMicrophones, 1);
            //LAPACK with HARK_Matrix// R.eigenValues(EV,V);
            //LAPACK with HARK_Matrix// for(size_t ii = 0; ii < NumMicrophones; ii++){
            //LAPACK with HARK_Matrix//     EigenValue[k][ii] = EV(ii, 0).real();
            //LAPACK with HARK_Matrix// }
            //LAPACK with HARK_Matrix// sort_eig_fvec_ascending(EigenValue[k], V, EigenValue[k], V);
            //LAPACK with HARK_Matrix// //sort_eig_fvec_descending(EigenValue[k], V, EigenValue[k], V);
            // boost with evd.h impl // //evd(R, EigenValue[k], V);


            Matrix<std::complex<float> > EV;
            Eigen::ComplexEigenSolver<Eigen::MatrixXcf> es(R);
            EV = es.eigenvalues();
            V = es.eigenvectors();
            std::vector<std::pair<float, size_t> > D;
            D.reserve(EV.size());
            for(size_t i = 0; i < EV.size(); i++){
                D.push_back(std::make_pair(EV.coeff(i, 0).real(), i));
            }
            // ascending order
            std::sort(D.begin(), D.end());
            // descending order
            //std::sort(D.begin(), D.end(), [](std::pair<float, size_t>& a, std::pair<float, size_t>& b){ return a.first > b.first; });
            Matrix<std::complex<float> > sortedV;
            sortedV.resizeLike(V);
            for(size_t i = 0; i < EV.size(); i++){
                //EV.coeffRef(i, 0) = D[i].first;
                EigenValue[k][i] = D[i].first;
                sortedV.col(i) = V.col(D[i].second);
            }
            V.noalias() = sortedV;
        }
        else if(music_alg == "GEVD") {
            //LAPACK with HARK_Matrix// //std::cout << "RN:\n" << RN << std::endl;
            //LAPACK with HARK_Matrix// TempCalc1 = RN.Inverse();
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc1 = RN.Inverse():\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc1:\n" << TempCalc1 << std::endl;
            //LAPACK with HARK_Matrix// Matrix<std::complex<float> > EV(NumMicrophones, 1);
            //LAPACK with HARK_Matrix// //std::cout << "EV(NumMicrophone, 1):\n"
            //LAPACK with HARK_Matrix// //          << "  EV:\n" << EV << std::endl;
            //LAPACK with HARK_Matrix// TempCalc1.eigenValues(EV, VN);
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc1.eigenValues(EV, VN):\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc1:\n" << TempCalc1 << "\n"
            //LAPACK with HARK_Matrix// //          << "  EV:\n" << EV << "\n"
            //LAPACK with HARK_Matrix// //          << "  VN:\n" << VN << std::endl;
            //LAPACK with HARK_Matrix// for(int ii = 0; ii < NumMicrophones; ii++){
            //LAPACK with HARK_Matrix//     REigenValue[ii] = EV(ii, 0).real();
            //LAPACK with HARK_Matrix// }
            //LAPACK with HARK_Matrix// sort_eig_fvec_ascending(REigenValue, VN, REigenValue, VN);//
            //LAPACK with HARK_Matrix// //sort_eig_fvec_descending(REigenValue, VN, REigenValue, VN);//
            //LAPACK with HARK_Matrix// //evd.hでは昇順で並べ替えられている
            //LAPACK with HARK_Matrix// //std::cout << "sort_eig_fvec_ascending(REigenValue, VN, REigenValue, VN):\n"
            //LAPACK with HARK_Matrix// //          << "  VN:\n" << VN << std::endl;
            // boost with evd.h impl // //invMAT(RN, TempCalc1);
            // boost with evd.h impl // //evd(TempCalc1, REigenValue, VN);

            // LAPACK and boost impl // //for(size_t i = 0; i < REigenValue.size(); i++) {
            // LAPACK and boost impl // //    if(REigenValue[i] < 0) {
            // LAPACK and boost impl // //        REigenValue[i] = -1.0f * std::pow(-1.0f * REigenValue[i], 0.5f);
            // LAPACK and boost impl // //    }
            // LAPACK and boost impl // //    else {
            // LAPACK and boost impl // //        REigenValue[i] = std::pow(REigenValue[i], 0.5f);
            // LAPACK and boost impl // //    }
            // LAPACK and boost impl // //}

            //LAPACK with HARK_Matrix// TempCalc2 = VN.Inverse();
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc2 = VN.Inverse():\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc2:\n" << TempCalc2 << std::endl;
            //LAPACK with HARK_Matrix// for(size_t i = 0; i < TempCalc3.rows(); i++) {
            //LAPACK with HARK_Matrix//     for(size_t j = 0; j < TempCalc3.cols(); j++) {
            //LAPACK with HARK_Matrix//         if(i == j) {
            //LAPACK with HARK_Matrix//             TempCalc3(i, j) = REigenValue[i];
            //LAPACK with HARK_Matrix//         }
            //LAPACK with HARK_Matrix//         else {
            //LAPACK with HARK_Matrix//             TempCalc3(i, j) = 0.0f;
            //LAPACK with HARK_Matrix//         }
            //LAPACK with HARK_Matrix//     }
            //LAPACK with HARK_Matrix// }
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc3 = REigenValue.asDiagonal():\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc3:\n" << TempCalc3 << std::endl;
            //LAPACK with HARK_Matrix// TempCalc6.noalias() = TempCalc3 * TempCalc2;
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc6 = TempCalc3 * TempCalc2:\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc6:\n" << TempCalc6 << std::endl;
            //LAPACK with HARK_Matrix// TempCalc4.noalias() = VN * TempCalc6;
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc4 = VN * TempCalc6:\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc4:\n" << TempCalc4 << std::endl;
            //LAPACK with HARK_Matrix// //std::cout << "R:\n" << R << std::endl;
            //LAPACK with HARK_Matrix// TempCalc7.noalias() = R * TempCalc4;
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc7 = R * TempCalc4:\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc7:\n" << TempCalc7 << std::endl;
            //LAPACK with HARK_Matrix// TempCalc5.noalias() = TempCalc4 * TempCalc7;
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc5 = TempCalc4 * TempCalc7:\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc5:\n" << TempCalc5 << std::endl;
            //LAPACK with HARK_Matrix//
            //LAPACK with HARK_Matrix// Matrix<std::complex<float> > EVtemp(NumMicrophones,1);
            //LAPACK with HARK_Matrix// //std::cout << "EVtemp(NumMicrophone, 1):\n"
            //LAPACK with HARK_Matrix// //          << "  EVtemp:\n" << EVtemp << std::endl;
            //LAPACK with HARK_Matrix// TempCalc5.eigenValues(EVtemp, V);
            //LAPACK with HARK_Matrix// //std::cout << "TempCalc5.eigenValues(EVtemp, V):\n"
            //LAPACK with HARK_Matrix// //          << "  TempCalc5:\n" << TempCalc5 << "\n"
            //LAPACK with HARK_Matrix// //          << "  EVtemp:\n" << EVtemp << "\n"
            //LAPACK with HARK_Matrix// //          << "  V:\n" << V << std::endl;
            //LAPACK with HARK_Matrix// for(size_t ii = 0; ii < NumMicrophones; ii++){
            //LAPACK with HARK_Matrix//     EigenValue[k][ii] = EVtemp(ii, 0).real();
            //LAPACK with HARK_Matrix// }
            //LAPACK with HARK_Matrix// sort_eig_fvec_ascending(EigenValue[k], V, EigenValue[k], V);//
            //LAPACK with HARK_Matrix// //sort_eig_fvec_descending(EigenValue[k], V, EigenValue[k], V);//

            // boost with evd.h impl // //invMAT(VN, TempCalc2);
            // boost with evd.h impl // //for(size_t i = 0; i < TempCalc3.size1(); i++) {
            // boost with evd.h impl // //    for(size_t j = 0; j < TempCalc3.size2(); j++) {
            // boost with evd.h impl // //        TempCalc3(i, j) = 0.0;
            // boost with evd.h impl // //        if(i == j) {
            // boost with evd.h impl // //            TempCalc3(i, j) = REigenValue[i];
            // boost with evd.h impl // //        }
            // boost with evd.h impl // //    }
            // boost with evd.h impl // //}
            // boost with evd.h impl // //noalias(TempCalc6) = prod(TempCalc3, TempCalc2);
            // boost with evd.h impl // //noalias(TempCalc4) = prod(VN, TempCalc6);
            // boost with evd.h impl // //noalias(TempCalc7) = prod(R, TempCalc4);
            // boost with evd.h impl // //noalias(TempCalc5) = prod(TempCalc4, TempCalc7);
            // boost with evd.h impl // //evd(TempCalc5, EigenValue[k], V);

            TempCalc1 = RN.inverse();
            Matrix<std::complex<float> > EV;
            Eigen::ComplexEigenSolver<Eigen::MatrixXcf> es(TempCalc1);
            EV = es.eigenvalues();
            VN = es.eigenvectors();
            std::vector<std::pair<float, size_t> > D;
            D.reserve(EV.size());
            for(size_t i = 0; i < EV.size(); i++){
                D.push_back(std::make_pair(EV.coeff(i, 0).real(), i));
            }
            // ascending order
            std::sort(D.begin(), D.end());
            // descending order
            //std::sort(D.begin(), D.end(), [](std::pair<float, size_t>& a, std::pair<float, size_t>& b){ return a.first > b.first; });
            Matrix<std::complex<float> > sortedVN;
            sortedVN.resizeLike(VN);
            for(size_t i = 0; i < EV.size(); i++){
                //EV.coeffRef(i, 0) = D[i].first;
                //REigenValue[i] = D[i].first;
                if(D[i].first < 0){
                    REigenValue[i] = -1.0f * std::pow(-1.0f * D[i].first, 0.5f);
                }
                else{
                    REigenValue[i] = std::pow(D[i].first, 0.5f);
                }
                sortedVN.col(i) = VN.col(D[i].second);
            }
            VN.noalias() = sortedVN;


            TempCalc2 = VN.inverse();
            TempCalc3.setZero();
            for(size_t i = 0; i < TempCalc3.rows(); i++){
                TempCalc3(i, i) = REigenValue[i];
            }
            //Eigen::VectorXd tmp_e = Eigen::Map<Eigen::VectorXd, Eigen::Unaligned>(REigenValue.data(), REigenValue.size());
            //TempCalc3.noalias() = tmp_e.asDiagonal().toDenseMatrix();

            //TempCalc6.noalias() = TempCalc3 * TempCalc2;
            //TempCalc4.noalias() = VN * TempCalc6;
            //TempCalc7.noalias() = R * TempCalc4;
            //TempCalc5.noalias() = TempCalc4 * TempCalc7;
            TempCalc4.noalias() = VN * (TempCalc3 * TempCalc2);
            TempCalc5.noalias() = TempCalc4 * (R * TempCalc4);

            Matrix<std::complex<float> > EVtemp;
            Eigen::ComplexEigenSolver<Eigen::MatrixXcf> es2(TempCalc5);
            EVtemp = es2.eigenvalues();
            V = es2.eigenvectors();
            std::vector<std::pair<float, size_t> > D2;
            D2.reserve(EVtemp.size());
            for(size_t i = 0; i < EVtemp.size(); i++){
                D2.push_back(std::make_pair(EVtemp.coeff(i, 0).real(), i));
            }
            // ascending order
            std::sort(D2.begin(), D2.end());
            // descending order
            //std::sort(D2.begin(), D2.end(), [](std::pair<float, size_t>& a, std::pair<float, size_t>& b){ return a.first > b.first; });
            Matrix<std::complex<float> > sortedV;
            sortedV.resizeLike(V);
            for(size_t i = 0; i < EVtemp.size(); i++){
                //EVtemp.coeffRef(i, 0) = D[i].first;
                EigenValue[k][i] = D2[i].first;
                sortedV.col(i) = V.col(D2[i].second);
            }
            V.noalias() = sortedV;
        }
        else if (music_alg == "GSVD") {
            //LAPACK with HARK_Matrix// TempCalc1 = RN.Inverse();
            //LAPACK with HARK_Matrix// TempCalc2 = TempCalc1 * R;
            //LAPACK with HARK_Matrix// //TempCalc2.singularValueDecomposition(TempCalc3,TempCalc5,TempCalc4);//u,s,v
            //LAPACK with HARK_Matrix// TempCalc2.singularValueDecomposition(TempCalc4,TempCalc5,TempCalc3);//u,s,v
            //LAPACK with HARK_Matrix// for(size_t ii=0; ii<NumMicrophones; ii++){
            //LAPACK with HARK_Matrix//     REigenValue[ii] = TempCalc5.getElementAt(ii,0).real();
            //LAPACK with HARK_Matrix// }
            //LAPACK with HARK_Matrix// sort_eig_fvec_ascending(REigenValue, TempCalc4, EigenValue[k], V);
            //LAPACK with HARK_Matrix// //sort_eig_fvec_descending(REigenValue, TempCalc4, EigenValue[k], V);

            // boost with evd.h impl // //invMAT(RN, TempCalc1);
            // boost with evd.h impl // //noalias(TempCalc2) = prod(TempCalc1, R);
            // boost with evd.h impl // //svd(TempCalc2, REigenValue, TempCalc4, TempCalc3);//a,s,u,v
            // boost with evd.h impl // //sort_eig_fvec_ascending(REigenValue, TempCalc4, EigenValue[k], V);

            TempCalc1 = RN.inverse();
            TempCalc2.noalias() = TempCalc1 * R;

            ////TempCalc2.singularValueDecomposition(TempCalc3,TempCalc5,TempCalc4);//u,s,v
            //TempCalc2.singularValueDecomposition(TempCalc4,TempCalc5,TempCalc3);//u,s,v

            Eigen::JacobiSVD<Eigen::MatrixXcf> svd(TempCalc2, Eigen::ComputeFullU | 
            Eigen::ComputeFullV);
            TempCalc3 = svd.matrixV();
            TempCalc4 = svd.matrixU();
            TempCalc5 = svd.singularValues();

            std::vector<std::pair<float, size_t> > D;
            D.reserve(TempCalc5.size());
            for(size_t i = 0; i < TempCalc5.size(); i++){
                D.push_back(std::make_pair(TempCalc5.coeff(i, 0).real(), i));
            }
            // ascending order
            std::sort(D.begin(), D.end());
            // descending order
            //std::sort(D.begin(), D.end(), [](std::pair<float, size_t>& a, std::pair<float, size_t>& b){ return a.first > b.first; });
            Matrix<std::complex<float> > sortedV;
            sortedV.resizeLike(V);
            for(size_t i = 0; i < TempCalc5.size(); i++){
                //TempCalc5.coeffRef(i, 0) = D[i].first;
                EigenValue[k][i] = D[i].first;
                sortedV.col(i) = TempCalc4.col(D[i].second);
            }
            V.noalias() = sortedV;
        }

        E[k].noalias() = V;
    }

    //printf("\n");

    CalcAveragePower(EigenValue, E, A, UsesFreq, NumSource, Pavg, Pavg_bin, MSSW, enable_eigenvalue_weight);
    for(size_t ih = 0; ih < SearchId.size(); ih++) {
        Pavg_sum[ih] = 10 * log10(Pavg[ih] + 1.0f);
    }
    if(maxnum_output_peaks < 0){
        maxnum_output_peaks = NumSource;
    }
    FoundSourceCount = SearchLocation(Pavg_sum, Pavg, maxnum_output_peaks, FoundId, mindeg, maxdeg);
#if ENABLE_TIMER
    end_time = std::chrono::system_clock::now();
    std::cout << DurationLog(end_time - start_time) << std::endl;
#endif

}

int LocalizationMUSIC::GetTFNumChannels(char* filename) {
    int nmic = -1;
////#ifdef ENABLE_HARKIO3
    harkio_TransferFunction *tf = NULL;
    tf = harkio_TransferFunction_fromFile(filename, (char*) "transfer function");
    if(tf == NULL){
        std::cerr << "LocalizationMUSIC. Failed to read tf file." << std::endl;
        return -1;
    }
    if(tf->loctfs == NULL){
        std::cerr << "LocalizationMUSIC. No localization matrix in tf file." << std::endl;
        harkio_TransferFunction_delete(&tf);
        return -1;
    }
    nmic = tf->mics->size;
    if(tf->loctfs[0]->rows != nmic){
        std::cerr << "LocalizationMUSIC. The number of mics does not match to the size of tf file." << std::endl;
        harkio_TransferFunction_delete(&tf);
        return -1;
    }
    harkio_TransferFunction_delete(&tf);
////#endif
    return nmic;
}


bool LocalizationMUSIC::SetTransferFunctionAll(TransferFunction tfData, std::vector<int>& ch) {

    // Get Configuration
    const hark::Config config = tfData.GetConfig();

    // Get TransferFunction
    const std::map<int, Matrix<std::complex<float> > > mpLoc = tfData.GetLocalization();
    //const Matrix<std::complex<float> >    &mat = mpLoc.second();


    int slength = config.nfft / 2 + 1;
    if(mpLoc.begin()->second.cols() != slength) {
        std::cerr << "LocalizationMUSIC. The number of bins does not match to the size of tf input." << std::endl;
        std::cerr << "slength = " << slength <<"   ncols = "<< mpLoc.begin()->second.cols() << std::endl;
        return false;
    }

    //// Check FFT Length
    //if (config.nfft != TimeLength) {
    //    std::cerr << "LocalizationMUSIC. The FFT Length does not match between params and tf input." << std::endl;
    //    return false;
    //}

    //// Check sampling rate
    //if (config.samplingRate != SamplingRate) {
    //    std::cerr << "LocalizationMUSIC. SamplingRate does not match between params and tf input." << std::endl;
    //    return false;
    //}
    
    // Get Sources
    const std::vector<Position> Sources = tfData.GetSources();
    int nsrc = Sources.size();
    const Neighbor NH = tfData.GetNeighbors();

    if(NH.ids.size() != nsrc) {
        std::cerr << "LocalizationMUSIC. The number of loctfs does not match to the size of nbrs." << std::endl;
        return false;
    }
    
    // Get Microphones
    const std::vector<Position> Mics = tfData.GetMicrophones();
    int nmic = Mics.size();
    if(mpLoc.begin()->second.rows() != nmic) {
        std::cerr << "LocalizationMUSIC. The number of mics does not match to the size of tf input." << std::endl;
        return false;
    }

    int nmic_selected = ch.size();
    for(size_t c = 0; c < nmic_selected; c++) {
        if(!(ch[c] >= 0 && ch[c] < nmic)) {
            std::cerr << "LocalizationMUSIC. Nonexistent channel is selected." << std::endl;
            return false;
        }
    }

    A.resize(nsrc);
    for(size_t iid = 0; iid < nsrc; iid++) {
        Matrix<std::complex<float> > Amat(slength, nmic_selected);
        A[iid] = Amat;
    }

    for(size_t iid = 0; iid < nsrc; iid++) {
        const Matrix<std::complex<float> > &mat = mpLoc.at(iid);
        for(size_t m = 0; m < nmic_selected; m++) {
            for(size_t k = 0; k < slength; k++) {
                A[iid](k, m) = mat(ch[m], k);
            }
        }
    }

    // Source Info
    AId.resize(nsrc);
    for(size_t iid = 0; iid < nsrc; iid++) {
        AId[iid] = Sources[iid].id;
        //AId[iid] = tf->poses->pos[iid]->id;
    }

    SearchId.resize(nsrc);
    for(size_t iid = 0; iid < nsrc; iid++) {
        SearchId[iid].resize(3);
        if(Sources[iid].sys == TF_Coordinate::Cartesian) {
            SearchId[iid][0] = Sources[iid].coord[0];
            SearchId[iid][1] = Sources[iid].coord[1];
            SearchId[iid][2] = Sources[iid].coord[2];
        }
        if(Sources[iid].sys == TF_Coordinate::Polar) {
            float azimuth_radian = harkio_Position_convToRadian(Sources[iid].coord[0]);
            float elevation_radian = harkio_Position_convToRadian(Sources[iid].coord[1]);
            float coef = Sources[iid].coord[2] * cosf(elevation_radian);
            SearchId[iid][0] = coef * cosf(azimuth_radian);
            SearchId[iid][1] = coef * sinf(azimuth_radian);
            SearchId[iid][2] = Sources[iid].coord[2] * sinf(elevation_radian);
        }
    }

    nbrs_ids.resize(NH.ids.size());
    for(size_t i = 0; i < NH.ids.size(); i++) {
        nbrs_ids[i] = NH.ids[i];
    }
    nbrs_numNeighbors.resize(NH.neighbors.size());
    for(size_t i = 0; i < NH.neighbors.size(); i++) {
        nbrs_numNeighbors[i] = NH.neighbors[i].size();
    }

    
    nbrs_neighbors.resize(NH.ids.size());
    for(size_t i = 0; i < NH.ids.size(); i++) {
        nbrs_neighbors[i].resize(nbrs_numNeighbors[i]);
        for(size_t j = 0; j < nbrs_numNeighbors[i]; j++) {
            nbrs_neighbors[i][j] = NH.neighbors[i][j];
        }
    }

    Pavg_sum.resize(SearchId.size());
    Pavg.resize(SearchId.size());


    Pavg_bin.resize(A[0].rows());
    for(size_t bins = 0; bins < A[0].rows(); bins++) {
        Pavg_bin[bins].resize(SearchId.size());
    }

    std::cerr << SearchId.size() << " srcs, " << A[0].cols() << " microphones, " << (A[0].rows() - 1) * 2 << " points" << std::endl;

    return true;
}

bool LocalizationMUSIC::SetTransferFunctionPertial(TransferFunction tfData, std::vector<int>& ch) {

    if(A.size() <= 0) {
        std::cerr << "LocalizationMUSIC. TransferFunction must be inputed to set the Pertial TransferFunction." << std::endl;
    }

    // Get TransferFunction
    const std::map<int, Matrix<std::complex<float> > > mpLoc = tfData.GetLocalization();

    // Check Bin size 
    size_t slength = A[0].rows();
    if(mpLoc.begin()->second.cols() != slength) {
        std::cerr << "LocalizationMUSIC. The number of bins does not match to the size of tf input (Pertial)." << std::endl;
        std::cerr << "slength = " << slength << "   ncols = " << mpLoc.begin()->second.cols() << std::endl;
        return false;
    }

    // Check Num of Microphones
    size_t nmic = A[0].cols();
    if(mpLoc.begin()->second.rows() != nmic) {
        std::cerr << "LocalizationMUSIC. The number of mics does not match to the size of tf input." << std::endl;
        return false;
    }

    size_t nmic_selected = ch.size();
    for(size_t c = 0; c < nmic_selected; c++) {
        if (!(ch[c] >= 0 && ch[c] < nmic)) {
            std::cerr << "LocalizationMUSIC. Nonexistent channel is selected." << std::endl;
            return false;
        }
    }

    for(auto itrm = mpLoc.begin(); itrm != mpLoc.end(); ++itrm) {
        // Find the index of A by using srcID
        auto itrv = std::find(AId.begin(), AId.end(), itrm->first);
        if(itrv == AId.end()) continue; // not find
        const int src_ind = std::distance(AId.begin(), itrv);
        for(size_t m = 0; m < nmic_selected; m++) {
            for(size_t k = 0; k < slength; k++) {
                A[src_ind](k, m) = itrm->second(ch[m], k);
            }
        }
    }
    return true;
}

bool LocalizationMUSIC::ReadTransferFunction(char* filename, std::vector<int>& ch) {
////#ifdef ENABLE_HARKIO3
    harkio_TransferFunction *tf = NULL;
    tf = harkio_TransferFunction_fromFile(filename, (char*) "transfer function");
    // tf = harkio_TransferFunction_new();
    // libharkio3_TransferFunction_readlibhakio2(&tf, filename, NULL);
    // ck_assert(tf != NULL);
    if(tf == NULL) {
        std::cerr << "LocalizationMUSIC. Failed to read tf file." << std::endl;
        return false;
    }

    if(tf->loctfs == NULL) {
        std::cerr << "LocalizationMUSIC. No localization matrix in tf file." << std::endl;
        harkio_TransferFunction_delete(&tf);
        return false;
    }
    int slength = tf->cfg->nfft / 2 + 1;
    if(tf->loctfs[0]->cols != slength) {
        std::cerr << "LocalizationMUSIC. The number of bins does not match to the size of tf file." << std::endl;
        harkio_TransferFunction_delete(&tf);
        return false;
    }

    int nsrc = tf->poses->size;
    // if ( tf->loctfs.size() != nsrc) {
    //   std::cerr << "LocalizationMUSIC. The number of loctfs does not match to the number of srcs." << std::endl;
    //   harkio_TransferFunction_delete(&tf);
    //   return false;
    // }
    if(tf->nbrs->size != nsrc) {
        std::cerr << "LocalizationMUSIC. The number of loctfs does not match to the size of nbrs." << std::endl;
        harkio_TransferFunction_delete(&tf);
        return false;
    }

    int nmic = tf->mics->size;
        if(tf->loctfs[0]->rows != nmic) {
        std::cerr << "LocalizationMUSIC. The number of mics does not match to the size of tf file." << std::endl;
        harkio_TransferFunction_delete(&tf);
        return false;
    }

    int nmic2 = ch.size();
    for(size_t c = 0; c < ch.size(); c++) {
        if(!(ch[c] >= 0 && ch[c] < nmic)) {
            std::cerr << "LocalizationMUSIC. Nonexistent channel is selected." << std::endl;
            harkio_TransferFunction_delete(&tf);
           return false;
        }
    }
    A.resize(nsrc);

    for(size_t iid = 0; iid < nsrc; iid++) {
        Matrix<std::complex<float> > Amat(slength, nmic2);
        A[iid] = Amat;
    }
    COMPLEX_TYPE complexValue;
    for(size_t iid = 0; iid < nsrc; iid++) {
        for(size_t m = 0; m < nmic2; m++) {
            for(size_t k = 0; k < slength; k++) {
                harkio_Matrix_getValueComplex(tf->loctfs[iid], ch[m], k, &complexValue);
                A[iid](k, m) = complexValue;
            }
        }
    }

    AId.resize(nsrc);
    for(size_t iid = 0; iid < nsrc; iid++) {
        AId[iid] = tf->poses->pos[iid]->id;
    }

    SearchId.resize(nsrc);
    for(size_t iid = 0; iid < nsrc; iid++) {
        SearchId[iid].resize(3);
        if(tf->poses->pos[iid]->sys == Coordinate::Cartesian) {
            SearchId[iid][0] = tf->poses->pos[iid]->coord[0];
            SearchId[iid][1] = tf->poses->pos[iid]->coord[1];
            SearchId[iid][2] = tf->poses->pos[iid]->coord[2];
        }
        if(tf->poses->pos[iid]->sys == Coordinate::Polar) {
            float azimuth_radian = harkio_Position_convToRadian(tf->poses->pos[iid]->coord[0]);
            float elevation_radian = harkio_Position_convToRadian(tf->poses->pos[iid]->coord[1]);
            float coef = tf->poses->pos[iid]->coord[2] * cosf(elevation_radian);
            SearchId[iid][0] = coef * cosf(azimuth_radian);
            SearchId[iid][1] = coef * sinf(azimuth_radian);
            SearchId[iid][2] = tf->poses->pos[iid]->coord[2] * sinf(elevation_radian);
        }
    }

    nbrs_ids.resize(tf->nbrs->size);
    for(size_t i = 0; i < tf->nbrs->size; i++) {
        nbrs_ids[i] = tf->nbrs->ids[i];
    }
    nbrs_numNeighbors.resize(tf->nbrs->size);
    for(size_t i = 0; i < tf->nbrs->size; i++) {
        nbrs_numNeighbors[i] = tf->nbrs->numNeighbors[i];
    }
    nbrs_neighbors.resize(tf->nbrs->size);
    for(size_t i = 0; i < tf->nbrs->size; i++) {
        nbrs_neighbors[i].resize(nbrs_numNeighbors[i]);
        for(size_t j = 0; j < nbrs_numNeighbors[i]; j++) {
            nbrs_neighbors[i][j] = tf->nbrs->neighbors[i][j];
        }
    }

    harkio_TransferFunction_delete(&tf);

////#endif  // #ifdef ENABLE_HARKIO3

    Pavg_sum.resize(SearchId.size());
    Pavg.resize(SearchId.size());
    Pavg_bin.resize(A[0].rows());
    for(size_t bins = 0; bins < A[0].rows(); bins++) {
        Pavg_bin[bins].resize(SearchId.size());
    }

    std::cerr << SearchId.size() << " srcs, " << A[0].cols() << " microphones, " << (A[0].rows() - 1) * 2 << " points" << std::endl;

    return true;
}

void LocalizationMUSIC::CalcAveragePower(std::vector<std::vector<float> >& w,
                                         std::vector<Matrix<std::complex<float> > >& E,
                                         std::vector<Matrix<std::complex<float> > >& A,
                                         std::vector<bool>& UsesFreq, int num_source,
                                         std::vector<float>& Pavg, std::vector<std::vector<float> >& Pavg_bin,
                                         std::vector<float>& Weight, bool enable_eigenvalue_weight) {
    int nh = Pavg.size();
    int num_mic = w[0].size();

    for(size_t ih = 0; ih < nh; ih++) {
        Pavg[ih] = 0.0f;
    }

    for(size_t k = 0; k < Pavg_bin.size(); k++) {
        for(size_t ih = 0; ih < Pavg_bin[0].size(); ih++) {
            Pavg_bin[k][ih] = 0.0f;
        }
    }

    float G = 0.0f;
    float norm_a = 0.0f;

    for(size_t k = 0; k < UsesFreq.size(); k++) {
        if(!UsesFreq[k]) {
            continue;
        }
        for(size_t ih = 0; ih < nh; ih++) {
            G = 0.0f;
            for(int m = num_mic - num_source - 1; m >= 0; m--) {
                G += std::norm(A[ih].row(k).adjoint().cwiseProduct(E[k].col(m)).sum());
                //G += norm(ublas::inner_prod(ublas::herm(ublas::row(A[ih], k)), ublas::column(E[k], m)));
            }
            //norm_a = ublas::inner_prod(ublas::herm(ublas::row(A[ih], k)), ublas::row(A[ih], k)).real();
            norm_a = A[ih].row(k).conjugate().cwiseProduct(A[ih].row(k)).sum().real();
            float EV_Weight = (enable_eigenvalue_weight ? sqrt(w[k][num_mic - 1]) : 1.0f);
            Pavg[ih] += (norm_a / G) * Weight[k] * EV_Weight;
            Pavg_bin[k][ih] = 10 * log10((norm_a / G * Weight[k]) + 1.0f);
            std::cerr << "";
        }
    }
}

int LocalizationMUSIC::SearchLocation(std::vector<float>& Pavg_sum, std::vector<float>& Pavg, int n_source,
                                      std::vector<int>& direction, int mindeg, int maxdeg) {
    int n_result;
    int nh = Pavg.size();
    std::vector<int> candidate_id;
    std::vector<float> candidate_val;

    candidate_id.reserve(nh);
    candidate_id.resize(0);
    candidate_val.reserve(nh);
    candidate_val.resize(0);
    SearchPeak(Pavg_sum, candidate_id, candidate_val, mindeg, maxdeg);

    std::vector<int> index;
    for(size_t i = 0; i < candidate_id.size(); i++) {
        index.push_back(i);
    }

    std::sort(index.begin(), index.end(), std::bind<bool>(less_for_vec(), candidate_val, std::placeholders::_2, std::placeholders::_1));

    std::vector<int> res_id;
    for(size_t i = 0; i < index.size(); i++) {
        res_id.push_back(candidate_id[index[i]]);
    }

    std::vector<float> res_val;
    for(size_t i = 0; i < index.size(); i++) {
        res_val.push_back(candidate_val[index[i]]);
    }

    if(n_source == 0) {
        n_result = res_id.size();
    }
    else if(res_id.size() < n_source) {
        n_result = res_id.size();
    }
    else {
        n_result = n_source;
    }

    direction.resize(0);
    for(size_t i = 0; i < n_result; i++) {
        direction.push_back(res_id[i]);
    }

    return n_result;
}

void LocalizationMUSIC::SearchPeak(std::vector<float>& power, std::vector<int>& peak_cand_id, std::vector<float>& peak_cand_val,
                                   int mindeg, int maxdeg) {
    while(maxdeg - mindeg > 360) { maxdeg -= 360; }
    while(maxdeg - mindeg <= 0) { maxdeg += 360; }

    for(size_t ih = 0; ih < power.size(); ih++) {

        double x = (SearchId[ih][0] == -0.0) ? 0.0 : SearchId[ih][0];
        double y = (SearchId[ih][1] == -0.0) ? 0.0 : SearchId[ih][1];
        int deg = (int) (180.0 / M_PI * atan2(y, x));
        while(deg - mindeg > 360) { deg -= 360; }
        while(deg - mindeg <= 0) { deg += 360; }
        if((deg < mindeg) || (deg > maxdeg)) { continue; }
        float maxc = power[ih];
        bool pf = true;

        size_t i;
        for(i = 0; i < nbrs_ids.size(); i++) {
            if(nbrs_ids[i] == AId[ih]) {
                break;
            }
        }
        for(size_t j = 0; j < nbrs_numNeighbors[i]; j++) {
            size_t jh;
            for(jh = 0; jh < AId.size(); jh++) {
                if(nbrs_neighbors[i][j] == AId[jh]) {
                    break;
                }
            }
            if(maxc < power[jh]) {
                pf = false;
            }
        }
        if(pf) {
            peak_cand_id.push_back(ih);
            peak_cand_val.push_back(power[ih]);
        }
    }
}

} // end of namespace

/* end of file */
